@@ -1,92 +1,122 @@
 //! Provides an abstraction of several async runtimes
 //!
-//! This  allows OpenTelemetry to work with any current or future runtime. There are currently
-//! builtin implementations for [Tokio] and [async-std].
+//! This  allows OpenTelemetry to work with any current or future runtime. There is currently
+//! built-in implementation for [Tokio].
 //!
 //! [Tokio]: https://crates.io/crates/tokio
-//! [async-std]: https://crates.io/crates/async-std
 
-use futures_util::{future::BoxFuture, stream::Stream};
+use futures_util::stream::{unfold, Stream};
 use std::{fmt::Debug, future::Future, time::Duration};
 use thiserror::Error;
 
-/// A runtime is an abstraction of an async runtime like [Tokio] or [async-std]. It allows
-/// OpenTelemetry to work with any current and hopefully future runtime implementation.
+/// A runtime is an abstraction of an async runtime like [Tokio]. It allows
+/// OpenTelemetry to work with any current and hopefully future runtime implementations.
 ///
 /// [Tokio]: https://crates.io/crates/tokio
-/// [async-std]: https://crates.io/crates/async-std
+///
+/// # Note
+///
+/// OpenTelemetry expects a *multithreaded* runtime because its types can move across threads.
+/// For this reason, this trait requires the `Send` and `Sync` bounds. Single-threaded runtimes
+/// can implement this trait in a way that spawns the tasks on the same thread as the calling code.
+#[cfg(feature = "experimental_async_runtime")]
 pub trait Runtime: Clone + Send + Sync + 'static {
-    /// A future stream, which returns items in a previously specified interval. The item type is
-    /// not important.
-    type Interval: Stream + Send;
-
-    /// A future, which resolves after a previously specified amount of time. The output type is
-    /// not important.
-    type Delay: Future + Send + Unpin;
-
-    /// Create a [futures_util::stream::Stream], which returns a new item every
-    /// [std::time::Duration].
-    fn interval(&self, duration: Duration) -> Self::Interval;
-
     /// Spawn a new task or thread, which executes the given future.
     ///
     /// # Note
     ///
     /// This is mainly used to run batch span processing in the background. Note, that the function
     /// does not return a handle. OpenTelemetry will use a different way to wait for the future to
-    /// finish when TracerProvider gets shutdown. At the moment this happens by blocking the
+    /// finish when the caller shuts down.
+    ///
+    /// At the moment, the shutdown happens by blocking the
     /// current thread. This means runtime implementations need to make sure they can still execute
     /// the given future even if the main thread is blocked.
-    fn spawn(&self, future: BoxFuture<'static, ()>);
+    fn spawn<F>(&self, future: F)
+    where
+        F: Future<Output = ()> + Send + 'static;
 
-    /// Return a new future, which resolves after the specified [std::time::Duration].
-    fn delay(&self, duration: Duration) -> Self::Delay;
+    /// Return a future that resolves after the specified [Duration].
+    fn delay(&self, duration: Duration) -> impl Future<Output = ()> + Send + 'static;
+}
+
+/// Uses the given runtime to produce an interval stream.
+#[cfg(feature = "experimental_async_runtime")]
+#[allow(dead_code)]
+pub(crate) fn to_interval_stream<T: Runtime>(
+    runtime: T,
+    interval: Duration,
+) -> impl Stream<Item = ()> {
+    unfold((), move |_| {
+        let runtime_cloned = runtime.clone();
+
+        async move {
+            runtime_cloned.delay(interval).await;
+            Some(((), ()))
+        }
+    })
 }
 
 /// Runtime implementation, which works with Tokio's multi thread runtime.
-#[cfg(feature = "rt-tokio")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-tokio")))]
+#[cfg(all(feature = "experimental_async_runtime", feature = "rt-tokio"))]
+#[cfg_attr(
+    docsrs,
+    doc(cfg(all(feature = "experimental_async_runtime", feature = "rt-tokio")))
+)]
 #[derive(Debug, Clone)]
 pub struct Tokio;
 
-#[cfg(feature = "rt-tokio")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-tokio")))]
+#[cfg(all(feature = "experimental_async_runtime", feature = "rt-tokio"))]
+#[cfg_attr(
+    docsrs,
+    doc(cfg(all(feature = "experimental_async_runtime", feature = "rt-tokio")))
+)]
 impl Runtime for Tokio {
-    type Interval = tokio_stream::wrappers::IntervalStream;
-    type Delay = ::std::pin::Pin<Box<tokio::time::Sleep>>;
-
-    fn interval(&self, duration: Duration) -> Self::Interval {
-        crate::util::tokio_interval_stream(duration)
-    }
-
-    fn spawn(&self, future: BoxFuture<'static, ()>) {
+    fn spawn<F>(&self, future: F)
+    where
+        F: Future<Output = ()> + Send + 'static,
+    {
         #[allow(clippy::let_underscore_future)]
         // we don't have to await on the returned future to execute
         let _ = tokio::spawn(future);
     }
 
-    fn delay(&self, duration: Duration) -> Self::Delay {
-        Box::pin(tokio::time::sleep(duration))
+    fn delay(&self, duration: Duration) -> impl Future<Output = ()> + Send + 'static {
+        tokio::time::sleep(duration)
     }
 }
 
 /// Runtime implementation, which works with Tokio's current thread runtime.
-#[cfg(feature = "rt-tokio-current-thread")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-tokio-current-thread")))]
+#[cfg(all(
+    feature = "experimental_async_runtime",
+    feature = "rt-tokio-current-thread"
+))]
+#[cfg_attr(
+    docsrs,
+    doc(cfg(all(
+        feature = "experimental_async_runtime",
+        feature = "rt-tokio-current-thread"
+    )))
+)]
 #[derive(Debug, Clone)]
 pub struct TokioCurrentThread;
 
-#[cfg(feature = "rt-tokio-current-thread")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-tokio-current-thread")))]
+#[cfg(all(
+    feature = "experimental_async_runtime",
+    feature = "rt-tokio-current-thread"
+))]
+#[cfg_attr(
+    docsrs,
+    doc(cfg(all(
+        feature = "experimental_async_runtime",
+        feature = "rt-tokio-current-thread"
+    )))
+)]
 impl Runtime for TokioCurrentThread {
-    type Interval = tokio_stream::wrappers::IntervalStream;
-    type Delay = ::std::pin::Pin<Box<tokio::time::Sleep>>;
-
-    fn interval(&self, duration: Duration) -> Self::Interval {
-        crate::util::tokio_interval_stream(duration)
-    }
-
-    fn spawn(&self, future: BoxFuture<'static, ()>) {
+    fn spawn<F>(&self, future: F)
+    where
+        F: Future<Output = ()> + Send + 'static,
+    {
         // We cannot force push tracing in current thread tokio scheduler because we rely on
         // BatchSpanProcessor to export spans in a background task, meanwhile we need to block the
         // shutdown function so that the runtime will not finish the blocked task and kill any
@@ -102,34 +132,8 @@ impl Runtime for TokioCurrentThread {
         });
     }
 
-    fn delay(&self, duration: Duration) -> Self::Delay {
-        Box::pin(tokio::time::sleep(duration))
-    }
-}
-
-/// Runtime implementation, which works with async-std.
-#[cfg(feature = "rt-async-std")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-async-std")))]
-#[derive(Debug, Clone)]
-pub struct AsyncStd;
-
-#[cfg(feature = "rt-async-std")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-async-std")))]
-impl Runtime for AsyncStd {
-    type Interval = async_std::stream::Interval;
-    type Delay = BoxFuture<'static, ()>;
-
-    fn interval(&self, duration: Duration) -> Self::Interval {
-        async_std::stream::interval(duration)
-    }
-
-    fn spawn(&self, future: BoxFuture<'static, ()>) {
-        #[allow(clippy::let_underscore_future)]
-        let _ = async_std::task::spawn(future);
-    }
-
-    fn delay(&self, duration: Duration) -> Self::Delay {
-        Box::pin(async_std::task::sleep(duration))
+    fn delay(&self, duration: Duration) -> impl Future<Output = ()> + Send + 'static {
+        tokio::time::sleep(duration)
     }
 }
 
@@ -138,6 +142,7 @@ impl Runtime for AsyncStd {
 ///
 /// [log]: crate::logs::BatchLogProcessor
 /// [span]: crate::trace::BatchSpanProcessor
+#[cfg(feature = "experimental_async_runtime")]
 pub trait RuntimeChannel: Runtime {
     /// A future stream to receive batch messages from channels.
     type Receiver<T: Debug + Send>: Stream<Item = T> + Send;
@@ -152,6 +157,7 @@ pub trait RuntimeChannel: Runtime {
 }
 
 /// Error returned by a [`TrySend`] implementation.
+#[cfg(feature = "experimental_async_runtime")]
 #[derive(Debug, Error)]
 pub enum TrySendError {
     /// Send failed due to the channel being full.
@@ -160,12 +166,13 @@ pub enum TrySendError {
     /// Send failed due to the channel being closed.
     #[error("cannot send message to batch processor as the channel is closed")]
     ChannelClosed,
-    /// Any other send error that isnt covered above.
+    /// Any other send error that isn't covered above.
     #[error(transparent)]
     Other(#[from] Box<dyn std::error::Error + Send + Sync + 'static>),
 }
 
 /// TrySend is an abstraction of `Sender` that is capable of sending messages through a reference.
+#[cfg(feature = "experimental_async_runtime")]
 pub trait TrySend: Sync + Send {
     /// The message that will be sent.
     type Message;
@@ -176,7 +183,10 @@ pub trait TrySend: Sync + Send {
     fn try_send(&self, item: Self::Message) -> Result<(), TrySendError>;
 }
 
-#[cfg(any(feature = "rt-tokio", feature = "rt-tokio-current-thread"))]
+#[cfg(all(
+    feature = "experimental_async_runtime",
+    any(feature = "rt-tokio", feature = "rt-tokio-current-thread")
+))]
 impl<T: Send> TrySend for tokio::sync::mpsc::Sender<T> {
     type Message = T;
 
@@ -188,8 +198,11 @@ impl<T: Send> TrySend for tokio::sync::mpsc::Sender<T> {
     }
 }
 
-#[cfg(feature = "rt-tokio")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-tokio")))]
+#[cfg(all(feature = "experimental_async_runtime", feature = "rt-tokio"))]
+#[cfg_attr(
+    docsrs,
+    doc(cfg(all(feature = "experimental_async_runtime", feature = "rt-tokio")))
+)]
 impl RuntimeChannel for Tokio {
     type Receiver<T: Debug + Send> = tokio_stream::wrappers::ReceiverStream<T>;
     type Sender<T: Debug + Send> = tokio::sync::mpsc::Sender<T>;
@@ -206,8 +219,17 @@ impl RuntimeChannel for Tokio {
     }
 }
 
-#[cfg(feature = "rt-tokio-current-thread")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-tokio-current-thread")))]
+#[cfg(all(
+    feature = "experimental_async_runtime",
+    feature = "rt-tokio-current-thread"
+))]
+#[cfg_attr(
+    docsrs,
+    doc(cfg(all(
+        feature = "experimental_async_runtime",
+        feature = "rt-tokio-current-thread"
+    )))
+)]
 impl RuntimeChannel for TokioCurrentThread {
     type Receiver<T: Debug + Send> = tokio_stream::wrappers::ReceiverStream<T>;
     type Sender<T: Debug + Send> = tokio::sync::mpsc::Sender<T>;
@@ -223,29 +245,3 @@ impl RuntimeChannel for TokioCurrentThread {
         )
     }
 }
-
-#[cfg(feature = "rt-async-std")]
-impl<T: Send> TrySend for async_std::channel::Sender<T> {
-    type Message = T;
-
-    fn try_send(&self, item: Self::Message) -> Result<(), TrySendError> {
-        self.try_send(item).map_err(|err| match err {
-            async_std::channel::TrySendError::Full(_) => TrySendError::ChannelFull,
-            async_std::channel::TrySendError::Closed(_) => TrySendError::ChannelClosed,
-        })
-    }
-}
-
-#[cfg(feature = "rt-async-std")]
-#[cfg_attr(docsrs, doc(cfg(feature = "rt-async-std")))]
-impl RuntimeChannel for AsyncStd {
-    type Receiver<T: Debug + Send> = async_std::channel::Receiver<T>;
-    type Sender<T: Debug + Send> = async_std::channel::Sender<T>;
-
-    fn batch_message_channel<T: Debug + Send>(
-        &self,
-        capacity: usize,
-    ) -> (Self::Sender<T>, Self::Receiver<T>) {
-        async_std::channel::bounded(capacity)
-    }
-}
